/**
 * PETALS - PETALS Services Platform. Copyright (C) 2006 EBM WebSourcing
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * 
 */

package org.ow2.petals.container.lifecycle;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.net.URLDecoder;
import java.util.ArrayList;
import java.util.List;

import javax.jbi.JBIException;
import javax.jbi.component.Bootstrap;
import javax.jbi.component.Component;
import javax.jbi.management.ComponentLifeCycleMBean;
import javax.jbi.management.InstallerMBean;
import javax.jbi.management.LifeCycleMBean;
import javax.jbi.messaging.MessagingException;
import javax.management.ObjectName;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.transaction.TransactionManager;

import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.fractal.fraclet.annotation.annotations.type.Contingency;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.easywsdl.extensions.wsdl4complexwsdl.api.WSDL4ComplexWsdlException;
import org.ow2.petals.communication.jndi.client.JNDIService;
import org.ow2.petals.container.ContainerService;
import org.ow2.petals.container.thread.BootstrapThread;
import org.ow2.petals.container.thread.ComponentLifeCycleThread;
import org.ow2.petals.jbi.component.context.ComponentContext;
import org.ow2.petals.jbi.component.context.ComponentContextImpl;
import org.ow2.petals.jbi.component.context.ComponentInitialContext;
import org.ow2.petals.jbi.component.context.InstallationContextImpl;
import org.ow2.petals.jbi.descriptor.original.generated.ClassLoaderDelegationType;
import org.ow2.petals.jbi.descriptor.original.generated.ComponentType;
import org.ow2.petals.jbi.descriptor.original.generated.Jbi;
import org.ow2.petals.jbi.descriptor.original.generated.Component.SharedLibrary;
import org.ow2.petals.jbi.management.Context;
import org.ow2.petals.jbi.management.ManagementException;
import org.ow2.petals.jbi.management.admin.AdminService;
import org.ow2.petals.jbi.messaging.endpoint.JBIServiceEndpointImpl;
import org.ow2.petals.jbi.messaging.endpoint.ServiceEndpoint;
import org.ow2.petals.jbi.messaging.registry.EndpointRegistry;
import org.ow2.petals.jbi.messaging.routing.RouterService;
import org.ow2.petals.jbi.messaging.routing.RoutingException;
import org.ow2.petals.kernel.api.server.PetalsException;
import org.ow2.petals.kernel.api.service.ServiceEndpoint.EndpointType;
import org.ow2.petals.kernel.configuration.ConfigurationService;
import org.ow2.petals.kernel.configuration.ContainerConfiguration;
import org.ow2.petals.kernel.server.FractalHelper;
import org.ow2.petals.system.classloader.ClassLoaderService;
import org.ow2.petals.system.classloader.PetalsClassLoader;
import org.ow2.petals.system.repository.RepositoryService;
import org.ow2.petals.system.systemstate.SystemStateService;
import org.ow2.petals.util.LoggingUtil;

/**
 * @author alouis
 * @author Olivier Fabre - EBM Websourcing
 * @author wjoseph - eBMWebsourcing
 * 
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.ow2.petals.container.lifecycle.InstallerInterface.class))
public class Installer implements InstallerMBean, InstallerInterface, ComponentContextCommunication {

    /**
     * Bootstrap classloader id prefix.
     */
    public static final String BS_PREFIX = "BS";

    /**
     * Component classloader id prefix.
     */
    public static final String COMP_PREFIX = "COMP";

    public static final String INSTALLED = "Installed";

    public static final String UNINSTALLED = "Uninstalled";

    @Monolog(name = "logger")
    protected Logger logger;

    private LoggingUtil log;

    /**
     * JBI-Management :: Admin Service
     */
    @Requires(name = "admin", signature = org.ow2.petals.jbi.management.admin.AdminService.class)
    protected AdminService adminService;

    /**
     * Platform :: Loader Service
     */
    @Requires(name = "classloader", signature = org.ow2.petals.system.classloader.ClassLoaderService.class)
    protected ClassLoaderService classloaderService;

    /**
     * JBI-Messaging :: Router Service
     */
    @Requires(name = "router", signature = org.ow2.petals.jbi.messaging.routing.RouterService.class)
    protected RouterService router;

    /**
     * Platform :: Repository Service
     */
    @Requires(name = "repository", signature = org.ow2.petals.system.repository.RepositoryService.class)
    protected RepositoryService repositoryService;

    /**
     * JBI-Messaging :: EndpointRegistry
     */
    @Requires(name = "endpoint", signature = org.ow2.petals.jbi.messaging.registry.EndpointRegistry.class)
    protected EndpointRegistry endpointRegistry;

    /**
     * Communication :: JNDIService
     */
    @Requires(name = "jndi", signature = org.ow2.petals.communication.jndi.client.JNDIService.class)
    protected JNDIService jndiService;

    /**
     * Container :: Container Service
     */
    @Requires(name = "container", signature = org.ow2.petals.container.ContainerService.class)
    protected ContainerService container;

    /**
     * JBI-Management :: TransactionManager Service
     */
    @Requires(name = "transactionManager", signature = javax.transaction.TransactionManager.class, contingency = Contingency.OPTIONAL)
    protected TransactionManager transactionManager;

    /**
     * Platform component :: SystemState Service
     */
    @Requires(name = "systemstate", signature = org.ow2.petals.system.systemstate.SystemStateService.class)
    protected SystemStateService recoverySrv;

    /**
     * The Configuration service fractal component
     */
    @Requires(name = "configuration", signature = org.ow2.petals.kernel.configuration.ConfigurationService.class)
    private ConfigurationService configurationService;

    /**
     * The Fractal component
     */
    @Service
    private org.objectweb.fractal.api.Component fractalComponent;

    /**
     * The container configuration
     */
    private ContainerConfiguration containerConfiguration;

    private Context context;

    private String state;

    private Jbi jbiDescriptor;

    private String installationRootPath;

    private String workingRootPath;

    /**
     * The consumer endpoint
     */
    private ServiceEndpoint address;

    private InitialContext initialContext;

    private InstallationContextImpl installContext;

    private PetalsClassLoader bootstrapClassLoader;

    private BootstrapThread bootstrapThread;

    private Object activitySynchronizer;

    private ComponentContext componentContext;

    private org.ow2.petals.jbi.descriptor.original.generated.Component componentDescription;

    private String componentName;

    private Bootstrap jbiBootstrap;

    /**
     * Create the Installer that will manage installation / uninstallation of
     * the JBI component. The related bootstrap is created and initialize here.
     */
    public Installer() {
        super();
        this.state = UNINSTALLED;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.container.lifecycle.InstallerInterface#init(java.util.Map)
     */
    @SuppressWarnings("deprecation")
    public void init(Context ctxt) throws IOException, JBIException, NamingException,
            NoSuchInterfaceException {

        this.context = ctxt;

        this.state = UNINSTALLED;

        this.activitySynchronizer = new Object();

        this.jbiDescriptor = this.context.getDescriptor();

        this.componentDescription = this.jbiDescriptor.getComponent();

        this.componentName = this.componentDescription.getIdentification().getName();

        this.installationRootPath = URLDecoder.decode(this.context.getInstallUrl().getPath());

        this.workingRootPath = this.repositoryService.getComponentWorkDirectory(this.componentName)
                .getAbsolutePath();

        this.address = new JBIServiceEndpointImpl();
        this.address.setType(EndpointType.CONSUMER);
        this.address.getLocation().setComponentName(this.componentName);
        this.address.getLocation().setContainerName(this.containerConfiguration.getName());
        this.address.getLocation().setSubdomainName(this.containerConfiguration.getSubdomainName());

        this.initialContext = this.createComponentInitialContext();

        // 1. create the component context
        this.componentContext = this.createComponentContext();

        // 2. create the bootstrap classloader
        this.bootstrapClassLoader = this.createBootstrapClassloader();

        // 3. create the bootstrap object
        this.jbiBootstrap = this.loadBootstrap(this.bootstrapClassLoader);

        // 5. create installation context
        this.installContext = new InstallationContextImpl(this.componentDescription,
                this.componentContext, true);

        // 6. init the Bootstrap thread that execute actions on the Bootstrap
        this.bootstrapThread = this.createBootstrapThread(this.bootstrapClassLoader,
                this.jbiBootstrap);

        // 7. init bootstrap
        this.initBootstrap(true);

    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.management.InstallerMBean#getInstallerConfigurationMBean()
     */
    public ObjectName getInstallerConfigurationMBean() throws JBIException {
        return this.bootstrapThread.getExtensionMBeanName();
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.management.InstallerMBean#getInstallRoot()
     */
    public String getInstallRoot() {
        return this.componentContext.getInstallRoot();
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.management.InstallerMBean#install()
     */
    public ObjectName install() throws JBIException {
        this.log.start();
        ObjectName componentMBeanName = null;

        try {
            if (UNINSTALLED.equals(this.state)) {
                synchronized (this.activitySynchronizer) {
                    // init the boostrap
                    try {
                        this.bootstrapThread.onInstall();
                    } finally {
                        this.bootstrapThread.cleanUp();
                    }

                    // create the component life cycle
                    componentMBeanName = this.createComponentLifeCycleMBean();
                    this.setState(INSTALLED);

                }
            } else {
                throw new JBIException("The Installer can not be installed in this state: "
                        + this.state);
            }
        } catch (JBIException e) {
            this.log.error(e.getMessage(), e);
            if (e.getCause() != null) {
                Throwable ex = e;
                StringBuilder msg = new StringBuilder(e.getMessage());
                while (ex.getCause() != null) {
                    ex = ex.getCause();
                    if (ex.getMessage() != null) {
                        msg.append("\nCaused by: ").append(e.getMessage());
                    }
                }
                e = new JBIException(msg.toString());
            }

            throw e;
        }

        this.log.end();

        return componentMBeanName;
    }

    /**
     * Uninstall the JBI component. Call the Bootstrap onUninstall() and
     * cleanup() methods, unregister the jbiComponent object,
     * 
     * @throws JBIException
     *             the component is already uninstalled
     */
    public void uninstall() throws JBIException {
        this.log.start();

        try {
            if (this.adminService.getServiceUnitsLifeCyclesForComponent(this.componentName).size() != 0) {
                throw new JBIException(
                        "Can not uninstall a component which holds deployed Service Unit(s)");
            }

            ComponentLifeCycleMBean componentLifeCycle = null;
            if (this.adminService.isEngine(this.componentName)) {
                componentLifeCycle = this.adminService
                        .getEngineComponentLifeCycleByName(this.componentName);
            } else {
                componentLifeCycle = this.adminService
                        .getBindingComponentLifeCycleByName(this.componentName);
            }

            if (componentLifeCycle == null) {
                throw new JBIException("The component '" + this.componentName
                        + "' can not be uninstalled as it is not installed!");
            }

            String componentState = componentLifeCycle.getCurrentState();
            if ((INSTALLED.equals(this.state))
                    && (LifeCycleMBean.SHUTDOWN.equals(componentState) || LifeCycleMBean.UNKNOWN
                            .equals(componentState))) {
                synchronized (this.activitySynchronizer) {
                    this.doUninstall();
                    this.setState(UNINSTALLED);
                }
            } else {
                throw new JBIException("The component '" + this.componentName
                        + "' can not be uninstalled in the Installer state: " + this.state
                        + " and Component state: " + componentState);
            }
        } catch (JBIException e) {
            this.log.error(e);
            throw e;
        }

        this.log.end();
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.management.InstallerMBean#isInstalled()
     */
    public boolean isInstalled() {
        return INSTALLED.equals(this.state);
    }

    /**
     * Get the unique component name of this component, as assigned by the
     * identification section of this component's installation descriptor.
     * 
     * @return the component name; must be non-null and non-empty.
     */
    public String getComponentName() {
        return this.componentName;
    }

    /*
     * @see org.ow2.petals.container.lifecycle.ComponentContextCommunication#
     * getAddress()
     */
    public ServiceEndpoint getAddress() {
        return this.address;
    }

    /*
     * @see org.ow2.petals.container.lifecycle.ComponentContextCommunication#
     * getEndpointRegistry()
     */
    public EndpointRegistry getEndpointRegistry() {
        return this.endpointRegistry;
    }

    /*
     * @see org.ow2.petals.container.lifecycle.ComponentContextCommunication#
     * getInitialContext()
     */
    public InitialContext getInitialContext() {
        return this.initialContext;
    }

    /*
     * @see org.ow2.petals.container.lifecycle.ComponentContextCommunication#
     * getInstallationRoot()
     */
    public String getInstallationRoot() {
        return this.installationRootPath;
    }

    /*
     * @see org.ow2.petals.container.lifecycle.ComponentContextCommunication#
     * getJBIDescriptor()
     */
    public Jbi getJBIDescriptor() {
        return this.jbiDescriptor;
    }

    /*
     * @see org.ow2.petals.container.lifecycle.ComponentContextCommunication#
     * getAdminService()
     */
    public AdminService getAdminService() {
        return this.adminService;
    }

    /*
     * @see org.ow2.petals.container.lifecycle.ComponentContextCommunication#
     * getRouterService()
     */
    public RouterService getRouterService() {
        return this.router;
    }

    /*
     * @see org.ow2.petals.container.lifecycle.ComponentContextCommunication#
     * getTransactionManagerService()
     */
    public TransactionManager getTransactionManagerService() {
        return this.transactionManager;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.container.lifecycle.ComponentContextCommunication#
     * getWorkspaceRoot()
     */
    public String getWorkspaceRoot() {
        return this.workingRootPath;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.container.lifecycle.ComponentContextCommunication#getLogger
     * ()
     */
    public Logger getLogger() {
        return this.logger;
    }

    /**
     * Get the JBI component
     * 
     * @return the {@link Component} instance
     */
    public Component getComponent() {
        return this.componentContext.getComponent();
    }

    /**
     * initialize the bootstrap
     * 
     * @param install
     * @throws JBIException
     */
    public void initBootstrap(boolean install) throws JBIException {
        this.installContext.setInstall(install);
        try {
            this.bootstrapThread.init(this.installContext);
        } catch (JBIException e) {
            this.bootstrapThread.cleanUp();
            throw e;
        }

    }

    /**
     * Start of the petals component
     * 
     * @throws PetalsException
     * @throws JBIException
     * 
     */
    @LifeCycle(on = LifeCycleType.START)
    protected void start() throws PetalsException, JBIException {
        this.log = new LoggingUtil(this.logger);
        this.log.start();

        this.containerConfiguration = this.configurationService.getContainerConfiguration();

        this.log.end();
    }

    /**
     * Stop of the petals component
     * 
     */
    @LifeCycle(on = LifeCycleType.STOP)
    protected void stop() {
        this.log.call();

        try {
            this.bootstrapThread.shutdownThread();
        } catch (JBIException e) {
            this.log.warning("Failed to stop the Bootstrap thread : " + e.getMessage());
        }

        this.logger = null;
    }

    /**
     * Set the <code>State</code> reference of the current state. An
     * <code>LifeCycleStateEventAbstract</code> is thrown.
     * 
     * @param state
     * @throws JBIException
     */
    private void setState(String state) throws JBIException {
        this.state = state;

        try {
            this.recoverySrv.updateComponentInstallationState(this.getComponentName(), state);
        } catch (Exception e) {
            throw new JBIException("Installer state can't be persisted for recovery", e);
        }
    }

    /**
     * Create the classLoader for Bootstrap
     * 
     * @return classLoader for Bootstrap
     */
    private PetalsClassLoader createBootstrapClassloader() {
        this.log.call();

        List<String> classpath = this.componentDescription.getBootstrapClassPath().getPathElement();
        // trim the list of class path elements
        for (int i = 0; i < classpath.size(); i++) {
            classpath.get(i).trim();
        }

        boolean parentFirst = !ClassLoaderDelegationType.SELF_FIRST
                .equals(this.componentDescription.getBootstrapClassLoaderDelegation());

        return this.createClassLoader(BS_PREFIX, classpath, parentFirst, null);
    }

    /**
     * Create a classloader with the specified parameters.
     * 
     * @param prefix
     *            bootstrap or component prefix
     * @param classpath
     *            classpath of the bootstrap/component
     * @param parentFirst
     *            delegation mode
     * @param slNameList
     *            list of the name of the required sharedLibraries
     * @return a classloader
     */
    private PetalsClassLoader createClassLoader(String prefix, List<String> classpath,
            boolean parentFirst, List<String> slNameList) {
        this.log.start();

        // prepare bootstrap and component classloader
        List<URL> urls = new ArrayList<URL>();

        try {
            urls.add(new File(this.installationRootPath).toURI().toURL());
        } catch (Exception e) {
            // simple do nothing - JBI Package has responsibility in
            // checking the archive contents against informations
            // contained in JBI Descriptor first.
            this.log.warning("Unexpected error", e);
        }
        URL[] baseUrls = urls.toArray(new URL[urls.size()]);

        PetalsClassLoader classLoader = null;
        try {
            classLoader = (PetalsClassLoader) this.classloaderService.createComponentClassLoader(
                    prefix + this.componentName, baseUrls, classpath, parentFirst, slNameList);
        } catch (PetalsException pe) {
            // just log when some error occur
            this.log.error("Failed to create bootstrap class loader" + pe.getMessage(), pe);
        }

        this.log.end();
        return classLoader;
    }

    /**
     * Create the classloader for the component. Be careful, we use the
     * classpath set in the InstallationContext. (during the installation phase,
     * the Bootstrap can modify the component's classpath by setting it in the
     * InstallationContext object.)
     * 
     * @return classloader for the component
     */
    @SuppressWarnings("unchecked")
    private PetalsClassLoader createComponentClassloader() {
        this.log.start();

        List<String> classpath = this.installContext.getClassPathElements();

        // Delegation model
        boolean parentFirst = !ClassLoaderDelegationType.SELF_FIRST
                .equals(this.componentDescription.getComponentClassLoaderDelegation());

        // Shared libraries used by this component
        List<SharedLibrary> slList = this.componentDescription.getSharedLibraryList();
        List slNameList = new ArrayList();
        if (slList != null) {
            for (SharedLibrary sl : slList) {
                slNameList.add(sl.getContent().trim() + "-" + sl.getVersion());
            }
        }

        PetalsClassLoader componentClassloader = this.createClassLoader(COMP_PREFIX, classpath,
                parentFirst, slNameList);

        this.log.end();
        return componentClassloader;
    }

    /**
     * Create the thread that will be used to process component installation
     * steps. The thread-classloader is set with the given paremeter.
     * 
     * @param bootstrap
     *            classloader
     * @param bootstrap
     *            instance
     * @return the BootstrapThread
     */
    private BootstrapThread createBootstrapThread(PetalsClassLoader classloader, Bootstrap bootstrap) {
        BootstrapThread it = new BootstrapThread(this.getComponentName(), bootstrap,
                this.containerConfiguration.getTaskTimeout(), this.log);
        it.setContextClassLoader(classloader);
        it.start();

        return it;
    }

    /**
     * Create the thread that will be used to process component life cycle
     * steps. The thread-classloader is set with the given paremeter.
     * 
     * @param classloader
     * @return the BootstrapThread
     */
    private ComponentLifeCycleThread createComponentLifeCycleThread(Component component,
            PetalsClassLoader classloader) {
        ComponentLifeCycleThread clt = new ComponentLifeCycleThread(component,
                this.containerConfiguration.getTaskTimeout(), this.log);
        clt.setContextClassLoader(classloader);
        clt.start();
        return clt;
    }

    /**
     * Register the installer engine/binding in the life cycle manager service
     * 
     * @return
     * @throws JBIException
     * @throws PetalsException
     */
    private ObjectName createComponentLifeCycleMBean() throws JBIException {
        this.log.start();

        ObjectName objectName = null;

        // 1. create component classloader
        PetalsClassLoader cl = this.createComponentClassloader();

        // 2. create the component object
        Component jbiComponent = this.loadComponent(cl);

        // 3. set the jbiComponent reference to the contextImpl
        this.componentContext.setComponent(jbiComponent);

        // 4. create the componentLifeCycle thread
        ComponentLifeCycleThread lifeCycleThread = this.createComponentLifeCycleThread(
                jbiComponent, cl);

        try {
            // 5. Creation of the lifecycle
            org.objectweb.fractal.api.Component lifecycle = this.container
                    .createComponentLifeCycle(this.componentDescription);

            // 6. Start the lifecycle
            FractalHelper.startComponent(lifecycle);

            // Initialization of the lifecycle
            ((ComponentLifeCycleInterface) lifecycle.getFcInterface("service")).init(
                    this.componentContext, lifeCycleThread, this.context);

            // 7. inform the router about the component installation
            this.router.addComponent(this.componentContext);

            // 8. register engine or binding
            ComponentLifeCycle componentLifecycle = (ComponentLifeCycle) lifecycle
                    .getFcInterface("/content");
            if (ComponentType.BINDING_COMPONENT.equals(this.componentDescription.getType())) {
                objectName = this.adminService.registerBindingComponent(componentLifecycle,
                        this.componentName);
            } else if (ComponentType.SERVICE_ENGINE.equals(this.componentDescription.getType())) {
                objectName = this.adminService.registerEngineComponent(componentLifecycle,
                        this.componentName);
            }
        } catch (RoutingException e) {
            throw new JBIException(e);
        } catch (NoSuchInterfaceException e) {
            throw new JBIException(e);
        } catch (IllegalLifeCycleException e) {
            throw new JBIException(e);
        } catch (PetalsException e) {
            throw new JBIException(e);
        }

        this.log.end();
        return objectName;
    }

    /**
     * Create and return the component context
     * 
     * @return the component context
     * @throws NoSuchInterfaceException
     * @throws WSDL4ComplexWsdlException
     * @throws PetalsException
     */
    private ComponentContext createComponentContext() throws NoSuchInterfaceException {

        ComponentContext componentContextImpl = new ComponentContextImpl(
                (ComponentContextCommunication) this.fractalComponent.getFcInterface("/content"));

        this.context.setComponentContext(componentContextImpl);
        return componentContextImpl;
    }

    /**
     * The initial JNDI context for all component. If this initial context is
     * null, it will be created
     * 
     * @return the initial JNDI context for all component. If this initial
     *         context is null, it will be created
     * @throws PetalsException
     *             if {@link InitialContext} can't be created
     */
    private InitialContext createComponentInitialContext() throws NamingException {
        ComponentInitialContext componentInitialContext = null;
        componentInitialContext = new ComponentInitialContext(this.jndiService.getUsersContext());

        return componentInitialContext;
    }

    /**
     * Unregister the installer engine/binding from the life cycle manager
     * service
     * 
     * @throws ManagementException
     */
    private void unregisterComponentLifeCycleMBean() throws ManagementException {
        if (this.componentDescription != null) {
            if (ComponentType.BINDING_COMPONENT.equals(this.componentDescription.getType())) {
                this.adminService.unregisterBindingComponent(this.componentName);
            } else if (ComponentType.SERVICE_ENGINE.equals(this.componentDescription.getType())) {
                this.adminService.unregisterEngineComponent(this.componentName);
            }
        }
    }

    /**
     * load the bootstrap, with its own classloader
     * 
     * @return
     */
    private Bootstrap loadBootstrap(PetalsClassLoader classloader) throws JBIException {
        this.log.call();

        Class<?> bootstrapClass;
        Bootstrap jbiBootstrapTmp = null;
        try {
            bootstrapClass = classloader.loadClass(this.componentDescription
                    .getBootstrapClassName().trim());

            if (bootstrapClass == null) {
                throw new JBIException("Could not find bootstrap class: "
                        + this.componentDescription.getBootstrapClassName());
            }
            jbiBootstrapTmp = (Bootstrap) bootstrapClass.newInstance();

        } catch (ClassNotFoundException e) {
            throw new JBIException("Class not found: "
                    + this.componentDescription.getBootstrapClassName(), e);
        } catch (InstantiationException e) {
            throw new JBIException("Could not instantiate : "
                    + this.componentDescription.getBootstrapClassName(), e);
        } catch (IllegalAccessException e) {
            throw new JBIException("Illegal access on: "
                    + this.componentDescription.getBootstrapClassName(), e);
        }
        return jbiBootstrapTmp;
    }

    /**
     * Load the component, within its own classloader
     * 
     * @return the component
     */
    private Component loadComponent(PetalsClassLoader classloader) throws JBIException {
        this.log.call();
        Component jbiComponent;

        Class<?> componentClass;
        try {
            componentClass = classloader.loadClass(this.componentDescription
                    .getComponentClassName().getContent().trim());

            if (componentClass == null) {
                throw new JBIException("Could not find component class: "
                        + this.componentDescription.getComponentClassName());
            }

            jbiComponent = (Component) componentClass.newInstance();

        } catch (ClassNotFoundException e) {
            throw new JBIException("Class not found: "
                    + this.componentDescription.getComponentClassName(), e);
        } catch (InstantiationException e) {
            throw new JBIException("Could not instantiate : "
                    + this.componentDescription.getComponentClassName(), e);
        } catch (IllegalAccessException e) {
            throw new JBIException("Illegal access on: "
                    + this.componentDescription.getComponentClassName(), e);
        } catch (SecurityException e) {
            throw new JBIException("Security exception for java.util.logging.Logger", e);
        }

        return jbiComponent;
    }

    /**
     * Do the bootstrap uninstallation. All the exceptions occurred are logged.<br />
     * If errors occurs, only the first exception will be thrown.
     * 
     * @throws JBIException
     *             : the first error generated
     */
    private void doUninstall() {
        this.log.start();

        // Ensure that the DeliveryChannel of the component is closed
        // (if the component is in an unknown state, for instance)
        try {
            if ((this.componentContext.getDeliveryChannel() != null)
                    && this.componentContext.getDeliveryChannel().isOpened()) {
                this.log.debug("close the component DeliveryChannel");
                this.componentContext.getDeliveryChannel().close();
            }
        } catch (MessagingException e) {
            this.log.error("Unable to close the delivery channel: ", e);
        }
        // call uninstall on bootstrap
        try {
            this.bootstrapThread.onUninstall();
        } catch (JBIException e) {
            this.log.error("Unable to uninstall the bootstap: ", e);
        }

        // clean the bootstrap thread
        try {
            this.bootstrapThread.cleanUp();
        } catch (JBIException e) {
            this.log.error("Unable to clean up the bootstap: ", e);
        }

        // unregister the component life-cycle mbean
        try {
            this.unregisterComponentLifeCycleMBean();
        } catch (ManagementException e) {
            this.log.error("Unable to unregister the component lifecycle: ", e);
        }
        // unregister the optional extension mbean
        try {
            if (this.componentContext.getComponent().getLifeCycle() != null) {
                ObjectName extName = this.componentContext.getComponent().getLifeCycle()
                        .getExtensionMBeanName();
                if (extName != null) {
                    this.adminService.getMBeanServer().unregisterMBean(extName);
                }
            }
        } catch (Exception e) {
            this.log.error("Unable to unregister the extension of the component: ", e);
        }
        // inform the router about the component uninstallation
        try {
            this.router.removeComponent(this.componentContext);
        } catch (RoutingException e) {
            this.log.error("Unable to inform the router about the component uninstallation: ", e);
        }

        // remove the fractal component
        try {
            this.container.removeComponentLifeCycle(this.componentDescription);
        } catch (PetalsException e) {
            this.log.error("Unable to remove the fractal component: ", e);
        }

        // clean the component context
        this.componentContext.clear();

        // unload component and bootstrap class loaders
        this.classloaderService.deleteClassLoader(Installer.COMP_PREFIX + this.componentName);
        this.classloaderService.deleteClassLoader(Installer.BS_PREFIX + this.componentName);

        this.log.end();
    }

}
