/**
 * PETALS: PETALS Services Platform Copyright (C) 2005 EBM WebSourcing
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 * 
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: DeliveryChannelImpl.java,v 1.2 2005/07/22 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.messaging.exchange;

import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.nio.charset.Charset;
import java.util.MissingResourceException;
import java.util.logging.Level;

import javax.jbi.JBIException;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.transaction.SystemException;
import javax.xml.namespace.QName;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.commons.stream.InputStreamForker;
import org.ow2.petals.commons.stream.ReaderInputStream;
import org.ow2.petals.commons.threadlocal.Transformers;
import org.ow2.petals.jbi.component.context.ComponentContext;
import org.ow2.petals.jbi.management.transaction.TransactionManagerImpl;
import org.ow2.petals.jbi.messaging.routing.RoutingException;
import org.ow2.petals.util.LoggingUtil;

/**
 * 
 * @author Adrien LOUIS - EBM WebSourcing
 * @author Gael BLONDELLE - EBM WebSourcing
 * @author ofabre - EBM Websourcing
 * @author rnaudin - EBM Websourcing
 */
public class DeliveryChannelImpl implements DeliveryChannel {

    protected ComponentContext componentContext;

    protected LoggingUtil log;

    protected Logger logger;

    /**
     * This messageExchangeFactory is used to return
     * <code>MessageExchangeDecorator</code> on accept() methods.
     */
    protected MessageExchangeFactoryImpl messageExchangeFactory;

    protected boolean opened = true;

    /**
     * Constructor
     * 
     * @param context
     * @param router
     * @param logger
     */
    public DeliveryChannelImpl(final ComponentContext context) {
        this.logger = context.getLogger();

        this.log = new LoggingUtil(this.logger);

        this.messageExchangeFactory = new MessageExchangeFactoryImpl(this, context.getAddress(),
                this.logger);

        this.componentContext = context;
    }

    /**
     * Check if the <code>MessageExchange</code> is Ok (not null and not
     * terminated)
     * 
     * @throws MessagingException
     */
    private static final void checkMessageExchange(final MessageExchangeDecorator exchange)
            throws MessagingException {
        if (exchange == null) {
            throw new MessagingException("MessageExchange is null");
        }

        final MessageExchangeDecorator exchangeDecorator = (MessageExchangeDecorator) exchange;

        if (exchangeDecorator.getMessageExchange().isTerminated()) {
            throw new MessagingException(
                    "MessageExchange is terminated, it can not be sent any more");
        }
    }

    /*
     * (non-Javadoc)
     * @see javax.jbi.messaging.DeliveryChannel#accept()
     */
    public MessageExchange accept() throws MessagingException {
        this.log.call();
        final MessageExchange me = this.poll(-1);

        return me;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.messaging.DeliveryChannel#accept(long)
     */
    public MessageExchange accept(final long timeoutMS) throws MessagingException {
        this.log.call(timeoutMS);
        if (timeoutMS < 0) {
            throw new MessagingException("Timeout must be positive");
        }
        final MessageExchange me = this.poll(timeoutMS);
        return me;
    }

    /**
     * Check if the <code>DeliveryChannel</code> is opened
     * 
     * @throws MessagingException
     */
    public void checkDeliveryChannelIsOpened() throws MessagingException {
        if (!this.opened) {
            throw new MessagingException("DeliveryChannel is closed");
        }
    }

    /**
     * This method wake up all threads that are waiting for a message. This
     * method deregister all endpoints for this component.
     * 
     * @see javax.jbi.messaging.DeliveryChannel#close()
     */
    public synchronized void close() throws MessagingException {
        this.log.start();

        this.checkDeliveryChannelIsOpened();

        this.opened = false;

        try {
            this.componentContext.deregisterAllEndpoints();
        } catch (final JBIException e) {
            throw new MessagingException(e);
        }

        this.log.end();
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.messaging.DeliveryChannel#createExchangeFactory()
     */
    public MessageExchangeFactory createExchangeFactory() {
        this.log.call();

        final MessageExchangeFactoryImpl mef = new MessageExchangeFactoryImpl(this,
                this.componentContext.getAddress(), this.logger);

        return mef;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * javax.jbi.messaging.DeliveryChannel#createExchangeFactory(javax.xml.namespace
     * .QName)
     */
    public MessageExchangeFactory createExchangeFactory(final QName interfaceName) {
        this.log.call(interfaceName);

        return new MessageExchangeFactoryImpl(this, this.componentContext.getAddress(),
                this.logger, interfaceName, null, null);
    }

    /*
     * (non-Javadoc)
     * 
     * @seejavax.jbi.messaging.DeliveryChannel#createExchangeFactory(javax.jbi.
     * servicedesc.ServiceEndpoint)
     */
    public MessageExchangeFactory createExchangeFactory(final ServiceEndpoint endpoint) {
        this.log.call(endpoint);

        return new MessageExchangeFactoryImpl(this, this.componentContext.getAddress(),
                this.logger, null, null, endpoint);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * javax.jbi.messaging.DeliveryChannel#createExchangeFactoryForService(javax
     * .xml.namespace.QName)
     */
    public MessageExchangeFactory createExchangeFactoryForService(final QName serviceName) {
        this.log.call(serviceName);

        return new MessageExchangeFactoryImpl(this, this.componentContext.getAddress(),
                this.logger, null, serviceName, null);
    }

    public boolean isOpened() {
        return this.opened;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * javax.jbi.messaging.DeliveryChannel#send(javax.jbi.messaging.MessageExchange
     * )
     */
    public void send(final MessageExchange exchange) throws MessagingException {
        this.log.call("Exchange Id:" + exchange.getExchangeId());

        this.sendExchange((MessageExchangeDecorator) exchange, 0, false);
    }

    /*
     * (non-Javadoc)
     * 
     * @seejavax.jbi.messaging.DeliveryChannel#sendSync(javax.jbi.messaging.
     * MessageExchange)
     */
    public boolean sendSync(final MessageExchange exchange) throws MessagingException {
        return this.sendSync(exchange, 0);
    }

    /*
     * (non-Javadoc)
     * 
     * @seejavax.jbi.messaging.DeliveryChannel#sendSync(javax.jbi.messaging.
     * MessageExchange, long)
     */
    public boolean sendSync(final MessageExchange exchange, final long timeoutMS)
            throws MessagingException {
        this.log.call("Exchange Id:" + exchange.getExchangeId() + " - Timeout:" + timeoutMS);
        boolean isTimeout = false;

        // the exchange can be only a MessageExchangeDecorator
        final MessageExchangeDecorator exchangeDecorator = (MessageExchangeDecorator) exchange;

        try {
            this.sendExchange(exchangeDecorator, timeoutMS, true);
            if (exchangeDecorator.isTimeout()) {
                // reset the ownership of the ME
                // TODO do this in the NMR
                // reset the timeout
                exchangeDecorator.setTimeout(false);
                isTimeout = true;
            }
        } catch (final MessagingException e) {
            this.log.debug(e.getMessage());
            // get the ownership of the ME
            exchangeDecorator.setStatus(ExchangeStatus.ERROR);
            exchangeDecorator.getMessageExchange().setTerminated(true);

            // If the cause is an Interrupted exception, we return false
            Throwable causeException = e.getCause();
            while (causeException != null) {
                if (causeException instanceof InterruptedException) {
                    this.log.end();
                    return false;
                } else {
                    causeException = causeException.getCause();
                }
            }
            throw e;
        }

        this.log.end();

        return !isTimeout;
    }

    /**
     * Log the current message, it can be an Out, Fault or In message.
     * 
     * @param exchange
     * @throws TransformerException
     * @throws JBIException
     * @throws MissingResourceException
     */
    private void logMessage(final MessageExchangeDecorator exchange) {
        java.util.logging.Logger componentLogger = null;
        // dirty but will log to the component logger.
        try {
            componentLogger = this.componentContext.getLogger("", null);
            if (componentLogger.isLoggable(Level.FINEST)) {
                if (ExchangeStatus.ACTIVE.equals(exchange.getStatus())) {
                    NormalizedMessage normalizedMessage = exchange
                            .getMessage(MessageExchangeImpl.OUT_MSG);
                    String messageType = MessageExchangeImpl.OUT_MSG;
                    if (normalizedMessage == null) {
                        normalizedMessage = exchange.getFault();
                        messageType = "Fault";
                    }
                    if (normalizedMessage == null) {
                        normalizedMessage = exchange.getMessage(MessageExchangeImpl.IN_MSG);
                        messageType = MessageExchangeImpl.IN_MSG;
                    }

                    componentLogger.finest("Message " + messageType + " with Id '"
                            + exchange.getExchangeId() + "' to send");
                    componentLogger.finest("Operation : " + exchange.getOperation());
                    for (final Object propertyName : normalizedMessage.getPropertyNames()) {
                        componentLogger.finest("Message property '" + propertyName + "' : "
                                + normalizedMessage.getProperty((String) propertyName));
                    }

                    Source content = normalizedMessage.getContent();
                    if (content != null) {
                        if (content instanceof StreamSource) {
                            InputStreamForker streamForker;
                            final StreamSource streamContent = (StreamSource) content;
                            final InputStream isContent = streamContent.getInputStream();
                            if (isContent != null) {
                                // The StreamSource was created from an
                                // InputStream
                                streamForker = new InputStreamForker(isContent);
                            } else {
                                // The StreamSource was created from a Reader
                                // we wrap it as an InputStream
                                streamForker = new InputStreamForker(new ReaderInputStream(
                                        streamContent.getReader()));
                            }
                            normalizedMessage.setContent(new StreamSource(streamForker
                                    .getInputStreamOne()));
                            content = new StreamSource(streamForker.getInputStreamTwo());
                        }

                        final Transformer transformer = Transformers.getDefaultTransformer();
                        transformer.setOutputProperty(OutputKeys.ENCODING, Charset.defaultCharset()
                                .name());
                        final ByteArrayOutputStream byteArrayOutputStream = new ByteArrayOutputStream();
                        try {
                            transformer.transform(content, new StreamResult(byteArrayOutputStream));
                        } finally {
                            transformer.reset();
                        }
                        componentLogger.finest("Message Content : "
                                + new String(byteArrayOutputStream.toByteArray()));
                    } else {
                        componentLogger.finest("Message Content is null");
                    }

                    for (final Object attachmentName : normalizedMessage.getAttachmentNames()) {
                        componentLogger.finest("Got a message attachment named '" + attachmentName
                                + "'");
                    }
                } else if (ExchangeStatus.DONE.equals(exchange.getStatus())) {
                    componentLogger.finest("Message acknowledgement DONE with Id '"
                            + exchange.getExchangeId() + "' to send");
                } else {
                    componentLogger.finest("Message acknowledgement ERROR with Id '"
                            + exchange.getExchangeId() + "' to send");
                    componentLogger.finest("Message Error : " + exchange.getError().getMessage());
                }
            }
        } catch (final Throwable t) {
            if (componentLogger != null) {
                componentLogger.fine("failed to log the content of the message to send: "
                        + t.getMessage());
            }
        }
    }

    /**
     * Poll for a receipt message exchange.
     * 
     * @param timeoutMS
     * 
     * @return The received message exchange, or {@code null} if timeout
     *         occurred
     * 
     * @throws MessagingException
     */
    private MessageExchange poll(final long timeoutMS) throws MessagingException {
        this.log.start();

        this.checkDeliveryChannelIsOpened();

        MessageExchangeDecorator messageDecorator = null;
        try {
            messageDecorator = this.componentContext.getRouter().receive(this.componentContext,
                    timeoutMS);
        } catch (final RoutingException e) {
            throw new MessagingException(e);
        }

        if (messageDecorator != null && this.componentContext.getTransactionManager() != null) {
            // decode the transaction to the Message Exchange if it is
            // transacted
            TransactionManagerImpl.decodeTransactionContext(messageDecorator);
        }

        this.log.end();

        return messageDecorator;
    }

    /**
     * Send the exchange to the underlying layer.<br>
     * Unused data in the exchange are cleaned before.<br>
     * If the message to exchange is a mere status information, and the exchange
     * is set to not send acknowledgment, the message is not sent.<br>
     */
    private void sendExchange(final MessageExchangeDecorator exchangeDecorator, final long timeout,
            final boolean sync) throws MessagingException {
        this.checkDeliveryChannelIsOpened();

        DeliveryChannelImpl.checkMessageExchange(exchangeDecorator);

        this.logMessage(exchangeDecorator);

        // switch the classloader to the container one
        final ClassLoader componentClassLoader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(this.getClass().getClassLoader());
        try {
            // check if the exchange is terminated
            if (!ExchangeStatus.ACTIVE.equals(exchangeDecorator.getStatus())) {
                exchangeDecorator.setTerminated(true);
                if (sync) {
                    throw new MessagingException(
                            "Sending an acknowledgment in synchronous mode is forbidden");
                }
            }

            // encode the transaction to the Message Exchange if it is
            // transacted
            if (this.componentContext.getTransactionManager() != null) {
                TransactionManagerImpl.encodeTransactionContext(exchangeDecorator);
            }

            if (sync) {
                this.componentContext.getRouter().sendSync(this.componentContext,
                        exchangeDecorator, timeout);
                if (!exchangeDecorator.isTimeout()
                        && this.componentContext.getTransactionManager() != null) {
                    // decode the transaction to the Message Exchange if it is
                    // transacted
                    TransactionManagerImpl.decodeTransactionContext(exchangeDecorator);
                }
            } else {
                this.componentContext.getRouter().send(this.componentContext, exchangeDecorator);
            }

        } catch (final JBIException e) {
            throw new MessagingException(e);
        } catch (final SystemException e) {
            throw new MessagingException(e);
        } finally {
            Thread.currentThread().setContextClassLoader(componentClassLoader);
        }

        this.log.end();
    }
}
