/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005-20010 EBM WebSourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

package org.ow2.petals.jbi.messaging.routing.module;

import java.io.IOException;
import java.lang.management.ManagementFactory;
import java.lang.management.MemoryMXBean;
import java.lang.management.MemoryNotificationInfo;
import java.lang.management.MemoryPoolMXBean;
import java.lang.management.MemoryType;
import java.sql.SQLException;
import java.util.ArrayDeque;
import java.util.Deque;
import java.util.Iterator;
import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicLong;
import java.util.concurrent.locks.LockSupport;

import javax.management.Notification;
import javax.management.NotificationEmitter;
import javax.management.NotificationListener;

import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.jbi.component.context.ComponentContext;
import org.ow2.petals.jbi.messaging.endpoint.ServiceEndpoint;
import org.ow2.petals.jbi.messaging.exchange.MessageExchangeDecorator;
import org.ow2.petals.jbi.messaging.routing.RouterService;
import org.ow2.petals.jbi.messaging.routing.RoutingException;
import org.ow2.petals.jbi.messaging.routing.module.flow.PriorityModifier;
import org.ow2.petals.system.persistence.PersistenceService;
import org.ow2.petals.transport.util.TransportSendContext;
import org.ow2.petals.util.LoggingUtil;

/**
 * The module orderer the exchange according to their priority, proposing a
 * Priority Queue mechanism. It is possible to reassign a priority by using the
 * {@link PriorityModifier} Interface.
 * 
 * @author rnaudin - EBM Websourcing
 * @since 3.1
 */
@FractalComponent
@Provides(interfaces = { @Interface(name = "service", signature = SenderModule.class),
        @Interface(name = "prioritymodifier", signature = PriorityModifier.class) })
public class PriorityOrdererModule implements SenderModule, NotificationListener, PriorityModifier {

    /**
     * Inner class to process priorities.
     * 
     * @author rnaudin
     * 
     */
    private class PriorityProcessor extends Thread {

        public boolean isRunning = true;

        @Override
        public void run() {

            while (this.isRunning) {
                // wait incoming messages
                if (PriorityOrdererModule.this.pendingExchanges.get() == 0) {
                    if (PriorityOrdererModule.this.timeoutManagerTimer != null) {
                        PriorityOrdererModule.this.timeoutManagerTimer.cancel();
                        PriorityOrdererModule.this.timeoutManagerTimer = null;
                    }
                    LockSupport.park(this);
                }

                short counter = 3;
                MessageExchangeDecorator exchangeDecorator = null;

                // priority 3
                synchronized (PriorityOrdererModule.this.exchangePriority3) {
                    while (counter > 0) {
                        exchangeDecorator = PriorityOrdererModule.this.exchangePriority3.pollLast();
                        if (exchangeDecorator != null) {
                            synchronized (exchangeDecorator.getExchangeId()) {
                                exchangeDecorator.getExchangeId().notify();
                            }
                            PriorityOrdererModule.this.pendingExchanges.decrementAndGet();
                            counter--;
                        } else {
                            counter = 0;
                        }
                    }
                }

                // priority 2
                counter = 2;
                exchangeDecorator = null;
                synchronized (PriorityOrdererModule.this.exchangePriority2) {
                    while (counter > 0) {
                        exchangeDecorator = PriorityOrdererModule.this.exchangePriority2.pollLast();
                        if (exchangeDecorator != null) {
                            synchronized (exchangeDecorator.getExchangeId()) {
                                exchangeDecorator.getExchangeId().notify();
                            }
                            PriorityOrdererModule.this.pendingExchanges.decrementAndGet();
                            counter--;
                        } else {
                            counter = 0;
                        }
                    }
                }

                // priority 1
                exchangeDecorator = null;
                synchronized (PriorityOrdererModule.this.exchangePriority1) {
                    exchangeDecorator = PriorityOrdererModule.this.exchangePriority1.pollLast();
                    if (exchangeDecorator != null) {
                        synchronized (exchangeDecorator.getExchangeId()) {
                            exchangeDecorator.getExchangeId().notify();
                        }
                        PriorityOrdererModule.this.pendingExchanges.decrementAndGet();
                    }
                }
            }
        }

    }

    /**
     * Inner class to handle the timeout for exchanges pending in the 0 priority
     * queue.
     * 
     * @author rnaudin
     * 
     */
    private class TimeoutManager extends TimerTask {

        @Override
        public void run() {

            long currentTime = System.currentTimeMillis();

            for (Long timeout : PriorityOrdererModule.this.exchangeTimeouts.keySet()) {
                if (currentTime > timeout.longValue()) {
                    MessageExchangeDecorator exchangeDecorator = PriorityOrdererModule.this.exchangeTimeouts
                            .get(timeout);
                    exchangeDecorator.setTimeout(true);
                    PriorityOrdererModule.this.log.debug("Exchange with Id '"
                            + exchangeDecorator.getExchangeId() + "' is in send timeout");
                    try {
                        PriorityOrdererModule.this.modifyPriority(exchangeDecorator, (short) 3);
                    } catch (RoutingException e) {
                        // the exchange has been already moved from 0 priority
                        // queue, do nothing
                    }

                }

            }
        }

    }

    /**
     * Logger wrapper.
     */
    private LoggingUtil log;

    /**
     * The logger.
     */
    @Monolog(name = "logger")
    private Logger logger;

    /**
     * The Persistence Fractal Service
     */
    @Requires(name = "persistence", signature = PersistenceService.class)
    private PersistenceService persistenceService;

    /**
     * The priority property to set in message exchange.
     */
    public static final String PROPERTY_ROUTER_PRIORITY = "org.petals.ow2.router.priority";

    /**
     * The period to run the Timeout Manager task
     */
    public static final long TIMEOUT_MANAGER_PERIOD = 1000;

    /**
     * Waiting queue
     */
    private Deque<MessageExchangeDecorator> exchangePriority0;

    /**
     * Low priority queue
     */
    private Deque<MessageExchangeDecorator> exchangePriority1;

    /**
     * Medium priority queue
     */
    private Deque<MessageExchangeDecorator> exchangePriority2;

    /**
     * High priority queue
     */
    private Deque<MessageExchangeDecorator> exchangePriority3;

    /**
     * Map of exchange timeouts for the 0 priority queue
     */
    private Map<Long, MessageExchangeDecorator> exchangeTimeouts;

    /**
     * Pending exchanges in the priority queues
     */
    private AtomicLong pendingExchanges;

    /**
     * The priority processor
     */
    private PriorityProcessor priorityProcessor;

    /**
     * The timeout manager timer
     */
    private Timer timeoutManagerTimer;

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.jbi.messaging.routing.module.SenderModule#electEndpoints
     * (java.util.Map, org.ow2.petals.jbi.component.context.ComponentContext,
     * org.ow2.petals.jbi.messaging.exchange.MessageExchangeDecorator)
     */
    public void electEndpoints(Map<ServiceEndpoint, TransportSendContext> electedEndpoints,
            ComponentContext sourceComponentContext, MessageExchangeDecorator exchangeDecorator)
            throws RoutingException {

        String priority = (String) exchangeDecorator.getProperty(PROPERTY_ROUTER_PRIORITY);
        if (priority != null) {
            synchronized (exchangeDecorator.getExchangeId()) {
                if ("3".equals(priority)) {
                    synchronized (this.exchangePriority3) {
                        this.log.debug("Add Exchange with Id '" + exchangeDecorator.getExchangeId()
                                + "' to priority 3");
                        this.exchangePriority3.add(exchangeDecorator);
                    }
                } else if ("2".equals(priority)) {
                    synchronized (this.exchangePriority2) {
                        this.log.debug("Add Exchange with Id '" + exchangeDecorator.getExchangeId()
                                + "' to priority 2");
                        this.exchangePriority2.add(exchangeDecorator);
                    }
                } else if ("1".equals(priority)) {
                    synchronized (this.exchangePriority1) {
                        this.log.debug("Add Exchange with Id '" + exchangeDecorator.getExchangeId()
                                + "' to priority 1");
                        this.exchangePriority1.add(exchangeDecorator);
                    }
                } else if ("0".equals(priority)) {
                    synchronized (this.exchangePriority0) {
                        this.log.debug("Add Exchange with Id '" + exchangeDecorator.getExchangeId()
                                + "' to priority 0");
                        this.exchangePriority0.add(exchangeDecorator);
                        this.exchangeTimeouts.put((Long) exchangeDecorator
                                .getProperty(RouterService.PROPERTY_ROUTER_TIMETOLIVE),
                                exchangeDecorator);
                    }
                }

                if (this.pendingExchanges.getAndIncrement() == 0) {
                    this.timeoutManagerTimer = new Timer();
                    this.timeoutManagerTimer.schedule(new TimeoutManager(), TIMEOUT_MANAGER_PERIOD,
                            TIMEOUT_MANAGER_PERIOD);
                    LockSupport.unpark(this.priorityProcessor);
                }

                try {
                    // the exchange decorator instance monitor lock is already
                    // used by the transporter to handle synchronous send. we
                    // use the exchangeId instead
                    exchangeDecorator.getExchangeId().wait();
                } catch (InterruptedException e) {
                    if (this.exchangePriority3.remove(exchangeDecorator)
                            || this.exchangePriority2.remove(exchangeDecorator)
                            || this.exchangePriority1.remove(exchangeDecorator)
                            || this.exchangePriority0.remove(exchangeDecorator)) {
                        this.pendingExchanges.decrementAndGet();
                    }
                    throw new RoutingException("Message Exchange with Id '"
                            + exchangeDecorator.getExchangeId()
                            + "' has been interrupted from priority ordering", e);
                } finally {
                    exchangeDecorator.setProperty(PROPERTY_ROUTER_PRIORITY, null);
                }
            }

        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * javax.management.NotificationListener#handleNotification(javax.management
     * .Notification, java.lang.Object)
     */
    public void handleNotification(Notification notification, Object handback) {
        if (MemoryNotificationInfo.MEMORY_THRESHOLD_EXCEEDED.equals(notification.getType())) {
            this.log.info("Memory threshold reached, persist some pending exchange if any");
            int pendingMessage = this.exchangePriority0.size() + this.exchangePriority1.size()
                    + this.exchangePriority2.size() + this.exchangePriority3.size();
            int counter = 0;
            MessageExchangeDecorator exchangeDecorator = null;

            // target 50% of pending message to persist

            // first all the exchange in waiting state
            synchronized (this.exchangePriority0) {
                Iterator<MessageExchangeDecorator> iterator = this.exchangePriority0.iterator();
                while (iterator.hasNext()) {

                    try {
                        exchangeDecorator = iterator.next();
                        if (!exchangeDecorator.isMessageExchangeStored()) {
                            exchangeDecorator.persistExchange(this.persistenceService);
                            this.log.debug("Persist Message Exchange with Id '"
                                    + exchangeDecorator.getExchangeId() + "'");
                            counter++;
                        }
                    } catch (SQLException e) {
                        this.log.warning("Failed to persist Message Exchange with Id '"
                                + exchangeDecorator.getExchangeId() + "'", e);
                    } catch (IOException e) {
                        this.log.warning("Failed to store a monitored exchange with Id '"
                                + exchangeDecorator.getExchangeId() + "'", e);
                    }
                }
            }

            if (counter < pendingMessage / 2) {
                // persist the half of low priority
                synchronized (this.exchangePriority1) {
                    Iterator<MessageExchangeDecorator> iterator = this.exchangePriority1.iterator();
                    int limit = this.exchangePriority1.size() / 2;
                    while (iterator.hasNext() && limit > 0) {
                        try {
                            exchangeDecorator = iterator.next();
                            if (!exchangeDecorator.isMessageExchangeStored()) {
                                exchangeDecorator.persistExchange(this.persistenceService);
                                this.log.debug("Persist Message Exchange with Id '"
                                        + exchangeDecorator.getExchangeId() + "'");
                                counter++;
                                limit++;
                            }
                        } catch (SQLException e) {
                            this.log.warning("Failed to persist Message Exchange with Id '"
                                    + exchangeDecorator.getExchangeId() + "'", e);
                        } catch (IOException e) {
                            this.log.warning("Failed to persist Message Exchange with Id '"
                                    + exchangeDecorator.getExchangeId() + "'", e);
                        }
                    }
                }
            }

            if (counter < pendingMessage / 2) {
                // persist the half of medium priority
                synchronized (this.exchangePriority2) {
                    Iterator<MessageExchangeDecorator> iterator = this.exchangePriority2.iterator();
                    int limit = this.exchangePriority2.size() / 2;
                    while (iterator.hasNext() && limit > 0) {
                        try {
                            exchangeDecorator = iterator.next();
                            if (!exchangeDecorator.isMessageExchangeStored()) {
                                exchangeDecorator.persistExchange(this.persistenceService);
                                this.log.debug("Persist Message Exchange with Id '"
                                        + exchangeDecorator.getExchangeId() + "'");
                                counter++;
                                limit++;
                            }
                        } catch (SQLException e) {
                            this.log.warning("Failed to persist Message Exchange with Id '"
                                    + exchangeDecorator.getExchangeId() + "'", e);
                        } catch (IOException e) {
                            this.log.warning("Failed to persist Message Exchange with Id '"
                                    + exchangeDecorator.getExchangeId() + "'", e);
                        }
                    }
                }
            }

            if (counter < pendingMessage / 2) {
                // persist the half of high priority
                synchronized (this.exchangePriority3) {
                    Iterator<MessageExchangeDecorator> iterator = this.exchangePriority3.iterator();
                    int limit = this.exchangePriority3.size() / 2;
                    while (iterator.hasNext() && limit > 0) {
                        try {
                            if (!exchangeDecorator.isMessageExchangeStored()) {
                                exchangeDecorator = iterator.next();
                                exchangeDecorator.persistExchange(this.persistenceService);
                                this.log.debug("Persist Message Exchange with Id '"
                                        + exchangeDecorator.getExchangeId() + "'");
                                counter++;
                                limit++;
                            }
                        } catch (SQLException e) {
                            this.log.warning("Failed to persist Message Exchange with Id '"
                                    + exchangeDecorator.getExchangeId() + "'", e);
                        } catch (IOException e) {
                            this.log.warning("Failed to persist Message Exchange with Id '"
                                    + exchangeDecorator.getExchangeId() + "'", e);
                        }
                    }
                }
            }
        }

    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.jbi.messaging.routing.module.priority.PriorityModifier
     * #modifyPriority
     * (org.ow2.petals.jbi.messaging.exchange.MessageExchangeDecorator, short)
     */
    public void modifyPriority(MessageExchangeDecorator exchangeDecorator, short priority)
            throws RoutingException {
        this.log.call("Exchange Id '" + exchangeDecorator.getExchangeId() + "' - Priority "
                + priority);

        boolean found = false;

        synchronized (this.exchangePriority3) {
            found = this.exchangePriority3.remove(exchangeDecorator);
        }
        if (!found) {
            synchronized (this.exchangePriority2) {
                found = this.exchangePriority2.remove(exchangeDecorator);
            }
        }
        if (!found) {
            synchronized (this.exchangePriority1) {
                found = this.exchangePriority1.remove(exchangeDecorator);
            }
        }
        if (!found) {
            synchronized (this.exchangePriority0) {
                found = this.exchangePriority0.remove(exchangeDecorator);
                this.exchangeTimeouts.remove(exchangeDecorator);
            }
        }
        if (!found) {
            throw new RoutingException("The Message exchange with Id '"
                    + exchangeDecorator.getExchangeId() + "' is no more in priority queues");
        } else {
            if (priority == 3) {
                synchronized (this.exchangePriority3) {
                    this.exchangePriority3.add(exchangeDecorator);
                }
            } else if (priority == 2) {
                synchronized (this.exchangePriority2) {
                    this.exchangePriority2.add(exchangeDecorator);
                }
            } else if (priority == 1) {
                synchronized (this.exchangePriority1) {
                    this.exchangePriority1.add(exchangeDecorator);
                }
            } else if (priority == 0) {
                synchronized (this.exchangePriority0) {
                    this.exchangePriority0.add(exchangeDecorator);
                    this.exchangeTimeouts.put((Long) exchangeDecorator
                            .getProperty(RouterService.PROPERTY_ROUTER_TIMETOLIVE),
                            exchangeDecorator);
                }
            } else {
                throw new RoutingException("Priority " + priority + " is not supported");
            }
        }
    }

    /**
     * Start the Fractal component
     * 
     * @throws RoutingException
     */
    @LifeCycle(on = LifeCycleType.START)
    protected void start() {
        this.log = new LoggingUtil(this.logger);
        this.log.call();

        this.pendingExchanges = new AtomicLong();
        this.exchangePriority0 = new ArrayDeque<MessageExchangeDecorator>();
        this.exchangePriority1 = new ArrayDeque<MessageExchangeDecorator>();
        this.exchangePriority2 = new ArrayDeque<MessageExchangeDecorator>();
        this.exchangePriority3 = new ArrayDeque<MessageExchangeDecorator>();
        this.exchangeTimeouts = new ConcurrentHashMap<Long, MessageExchangeDecorator>();

        for (MemoryPoolMXBean memoryPoolMXBean : ManagementFactory.getMemoryPoolMXBeans()) {
            if (MemoryType.HEAP.equals(memoryPoolMXBean.getType())
                    && memoryPoolMXBean.isUsageThresholdSupported()) {
                long maxUsage = memoryPoolMXBean.getUsage().getMax();
                final Double threshold = new Double(maxUsage * 0.8);
                memoryPoolMXBean.setUsageThreshold(threshold.longValue());
            }
        }
        MemoryMXBean memoryMbean = ManagementFactory.getMemoryMXBean();
        ((NotificationEmitter) memoryMbean).addNotificationListener(this, null, null);

        this.priorityProcessor = new PriorityProcessor();
        this.priorityProcessor.start();
    }

    /**
     * Stop the Fractal component
     */
    @LifeCycle(on = LifeCycleType.STOP)
    protected void stop() {
        this.log.call();

        // the waiting threads should have been interrupted already by their
        // component owners, no need to treat these interruption here

        this.priorityProcessor.isRunning = false;
        LockSupport.unpark(this.priorityProcessor);
        if (this.timeoutManagerTimer != null) {
            this.timeoutManagerTimer.cancel();
        }
    }
}
