/**
 * PETALS - PETALS Services Platform. Copyright (c) 2009 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */

package org.ow2.petals.system.persistence;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.URISyntaxException;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;

import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.messaging.MessageExchange.Role;
import javax.xml.namespace.QName;

import org.apache.commons.dbcp.BasicDataSource;
import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.easywsdl.wsdl.api.abstractItf.AbsItfOperation;
import org.ow2.petals.commons.util.IDGenerator;
import org.ow2.petals.jbi.messaging.endpoint.ServiceEndpoint;
import org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl;
import org.ow2.petals.kernel.configuration.ConfigurationService;
import org.ow2.petals.kernel.configuration.ContainerConfiguration;
import org.ow2.petals.util.LoggingUtil;

/**
 * The Persistence Service.
 * <p>
 * It is using a JDBC pool of connection to persist Petals data.
 * 
 * @author rnaudin
 * @since 3.1
 * 
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = PersistenceService.class))
public class PersistenceServiceImpl implements PersistenceService {

    private class PersistenceCleaner extends TimerTask {

        @Override
        public void run() {

            PersistenceServiceImpl.this.log.debug("Start Persistence clean up");
            long endTime = System.currentTimeMillis() - PersistenceServiceImpl.this.storageDuration;

            try {
                final Connection connection = PersistenceServiceImpl.this.jdbcDataSource
                        .getConnection();
                final Statement getStatement = connection.createStatement();

                List<String> exchangeIds = new ArrayList<String>();
                try {
                    final ResultSet resultSet = getStatement
                            .executeQuery("SELECT EXCHANGE_ID FROM MESSAGE_EXCHANGE WHERE CREATION_TIMESTAMP < '"
                                    + endTime + "'");
                    while (resultSet.next()) {
                        exchangeIds.add(resultSet.getString(1));
                    }

                    for (String exchangeId : exchangeIds) {
                        PersistenceServiceImpl.this.log.debug("Delete Message Exchange with Id '"
                                + exchangeId + "' related data");
                        final Statement deleteStatement = connection.createStatement();
                        deleteStatement
                                .execute("DELETE FROM MESSAGE_EXCHANGE_PROPERTY WHERE EXCHANGE_ID='"
                                        + exchangeId + "'");
                        deleteStatement
                                .execute("DELETE FROM NORMALIZED_MESSAGE WHERE EXCHANGE_ID='"
                                        + exchangeId + "'");
                        deleteStatement
                                .execute("DELETE FROM EXCHANGE_TIMESTAMP WHERE EXCHANGE_ID='"
                                        + exchangeId + "'");
                        deleteStatement.execute("DELETE FROM MESSAGE_EXCHANGE WHERE EXCHANGE_ID='"
                                + exchangeId + "'");
                    }
                } finally {
                    connection.close();
                }
            } catch (SQLException e) {
                PersistenceServiceImpl.this.log.warning("Failed to delete old exchange data", e);
            }
        }
    }

    private Timer cleanerTimer;

    /**
     * The Configuration service fractal component
     */
    @Requires(name = "configuration", signature = org.ow2.petals.kernel.configuration.ConfigurationService.class)
    private ConfigurationService configurationService;

    private String databaseDriver;

    /**
     * The data source to handle JDBC connections pooling
     */
    private BasicDataSource jdbcDataSource;

    /**
     * logger wrapper
     */
    private LoggingUtil log;

    /**
     * Monolog Logger instance.
     */
    @Monolog(name = "logger")
    private Logger logger;

    private long storageDuration;

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.system.persistence.PersistenceService#addMessageExchange
     * (org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl, boolean,
     * java.lang.Boolean)
     */
    public void addMessageExchange(MessageExchangeImpl exchange, boolean storeExchange,
            Boolean isSent) throws SQLException, IOException {
        this.log.call("ExchangeId : " + exchange.getExchangeId());

        final Connection connection = this.jdbcDataSource.getConnection();
        
        long currentTime = System.currentTimeMillis();

        try {
            PreparedStatement storeStatement;
            if (!storeExchange) {
                storeStatement = connection
                        .prepareStatement("INSERT INTO MESSAGE_EXCHANGE (EXCHANGE_ID, CONSUMER_ENDPOINT, ENDPOINT, INTERFACE_NAME, SERVICE_NAME, ENDPOINT_NAME, OPERATION_NAME, PATTERN, CREATION_TIMESTAMP) VALUES(?,?,?,?,?,?,?,?,?)");
                storeStatement.setString(1, exchange.getExchangeId());
                storeStatement.setObject(2, exchange.getConsumerEndpoint());
                storeStatement.setObject(3, exchange.getServiceEndpoint());
                if (exchange.getInterfaceName() == null) {
                    storeStatement.setString(4, null);
                } else {
                    storeStatement.setString(4, exchange.getInterfaceName().toString());
                }
                if (exchange.getService() == null) {
                    storeStatement.setString(5, null);
                } else {
                    storeStatement.setString(5, exchange.getService().toString());
                }
                if (exchange.getServiceEndpoint() == null) {
                    storeStatement.setString(6, null);
                } else {
                    storeStatement.setString(6, exchange.getServiceEndpoint().getEndpointName());
                }
                if (exchange.getOperation() == null) {
                    storeStatement.setString(7, null);
                } else {
                    storeStatement.setString(7, exchange.getOperation().toString());
                }
                String mepString = null;
                // TODO remove that part when the deprecated MEP constant would
                // be removed from code
                if (MessageExchangeImpl.IN_ONLY_PATTERN.equals(exchange.getPattern())) {
                    mepString = AbsItfOperation.MEPPatternConstants.IN_ONLY.toString();
                } else if (MessageExchangeImpl.ROBUST_IN_ONLY_PATTERN.equals(exchange.getPattern())) {
                    mepString = AbsItfOperation.MEPPatternConstants.ROBUST_IN_ONLY.toString();
                } else if (MessageExchangeImpl.IN_OUT_PATTERN.equals(exchange.getPattern())) {
                    mepString = AbsItfOperation.MEPPatternConstants.IN_OUT.toString();
                } else if (MessageExchangeImpl.IN_OPTIONAL_OUT_PATTERN
                        .equals(exchange.getPattern())) {
                    mepString = AbsItfOperation.MEPPatternConstants.IN_OPTIONAL_OUT.toString();
                }
                storeStatement.setString(8, mepString);
                storeStatement.setLong(9, currentTime);
            } else {
                storeStatement = connection
                        .prepareStatement("INSERT INTO MESSAGE_EXCHANGE VALUES(?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)");

                storeStatement.setString(1, exchange.getExchangeId());
                storeStatement.setObject(2, exchange.getConsumerEndpoint());
                storeStatement.setObject(3, exchange.getServiceEndpoint());

                if (exchange.getInterfaceName() == null) {
                    storeStatement.setString(4, null);
                } else {
                    storeStatement.setString(4, exchange.getInterfaceName().toString());
                }
                if (exchange.getService() == null) {
                    storeStatement.setString(5, null);
                } else {
                    storeStatement.setString(5, exchange.getService().toString());
                }
                if (exchange.getServiceEndpoint() == null) {
                    storeStatement.setString(6, null);
                } else {
                    storeStatement.setString(6, exchange.getServiceEndpoint().getEndpointName());
                }
                if (exchange.getOperation() == null) {
                    storeStatement.setString(7, null);
                } else {
                    storeStatement.setString(7, exchange.getOperation().toString());
                }
                String mepString = null;
                // TODO remove that part when the deprecated MEP constant would
                // be removed from code
                if (MessageExchangeImpl.IN_ONLY_PATTERN.equals(exchange.getPattern())) {
                    mepString = AbsItfOperation.MEPPatternConstants.IN_ONLY.toString();
                } else if (MessageExchangeImpl.ROBUST_IN_ONLY_PATTERN.equals(exchange.getPattern())) {
                    mepString = AbsItfOperation.MEPPatternConstants.ROBUST_IN_ONLY.toString();
                } else if (MessageExchangeImpl.IN_OUT_PATTERN.equals(exchange.getPattern())) {
                    mepString = AbsItfOperation.MEPPatternConstants.IN_OUT.toString();
                } else if (MessageExchangeImpl.IN_OPTIONAL_OUT_PATTERN
                        .equals(exchange.getPattern())) {
                    mepString = AbsItfOperation.MEPPatternConstants.IN_OPTIONAL_OUT.toString();
                }
                storeStatement.setString(8, mepString);
                storeStatement.setObject(9, exchange.getError());
                if (Role.CONSUMER.equals(exchange.getRole())) {
                    storeStatement.setShort(10, MessageExchangeImpl.SERIALIZE_ROLE_CONSUMER);
                } else {
                    storeStatement.setShort(10, MessageExchangeImpl.SERIALIZE_ROLE_PROVIDER);
                }
                if (ExchangeStatus.ACTIVE.equals(exchange.getStatus())) {
                    storeStatement.setShort(11, MessageExchangeImpl.SERIALIZE_STATUS_ACTIVE);
                } else if (ExchangeStatus.DONE.equals(exchange.getStatus())) {
                    storeStatement.setShort(11, MessageExchangeImpl.SERIALIZE_STATUS_DONE);
                } else {
                    storeStatement.setShort(11, MessageExchangeImpl.SERIALIZE_STATUS_ERROR);
                }
                storeStatement.setBoolean(12, exchange.isTerminated());
                storeStatement.setBoolean(13, exchange.isTransacted());
                storeStatement.setBoolean(14, exchange.isMonitored());
                storeStatement.setBoolean(15, exchange.isPersisted());
                storeStatement.setLong(16, currentTime);
            }

            storeStatement.execute();

            if (storeExchange) {
                storeStatement = connection
                        .prepareStatement("INSERT INTO MESSAGE_EXCHANGE_PROPERTY VALUES(?,?,?,?)");
                for (Object key : exchange.getPropertyNames()) {
                    storeStatement.setString(1, exchange.getExchangeId() + (String) key);
                    storeStatement.setString(2, exchange.getExchangeId());
                    storeStatement.setString(3, (String) key);
                    if (HSQLDB_DRIVER.equals(this.databaseDriver)) {
                        storeStatement.setObject(4, exchange.getProperty((String) key));
                    } else {
                        // With MYSQL, the setObject() method serializes not
                        // properly primitive alter-ego Objects, use Streams
                        // instead
                        final ByteArrayOutputStream bytesOutputStream = new ByteArrayOutputStream();
                        new ObjectOutputStream(bytesOutputStream).writeUnshared(exchange
                                .getProperty((String) key));
                        ByteArrayInputStream bytesInputStream = new ByteArrayInputStream(
                                bytesOutputStream.toByteArray());
                        storeStatement.setBinaryStream(4, bytesInputStream, bytesInputStream
                                .available());
                    }

                    storeStatement.execute();
                }

                storeStatement = connection
                        .prepareStatement("INSERT INTO NORMALIZED_MESSAGE VALUES(?,?,?,?)");
                for (String key : exchange.getMessages().keySet()) {
                    storeStatement.setString(1, exchange.getExchangeId() + key);
                    storeStatement.setString(2, exchange.getExchangeId());
                    storeStatement.setString(3, key);
                    storeStatement.setObject(4, exchange.getMessage(key));

                    storeStatement.execute();
                }
            }

            if (isSent != null) {
                short isSentbinary = 0;
                if (isSent) {
                    isSentbinary = 1;
                }
                final String type;
                if (exchange.getMessage(MessageExchangeImpl.IN_MSG) != null) {
                    type = MessageExchangeImpl.IN_MSG;
                } else if (exchange.getMessage(MessageExchangeImpl.OUT_MSG) != null) {
                    type = MessageExchangeImpl.OUT_MSG;
                } else if (exchange.getMessage(MessageExchangeImpl.FAULT_MSG) != null) {
                    type = MessageExchangeImpl.FAULT_MSG;
                } else if (exchange.getError() != null) {
                    type = "error-ack";
                } else {
                    type = "done-ack";
                }

                String isSentValue = null;
                /*
                 * If Hsqldb use 'true'/'false' values
                 * for is_sent column, else use
                 * '1'/'0' values
                 */
                if(HSQLDB_DRIVER.equals(this.databaseDriver)){
                    if(isSentbinary == 0){
                        isSentValue = "false";
                    } else {
                        isSentValue = "true";
                    }
                } else {
                    isSentValue = String.valueOf(isSentbinary);
                }
                
                final Statement addStatement = connection.createStatement();
                addStatement
                        .execute("INSERT INTO EXCHANGE_TIMESTAMP (EVENT_TIMESTAMP, EXCHANGE_ID, IS_SENT, TYPE) VALUES('"
                                + currentTime
                                + "','"
                                + exchange.getExchangeId()
                                + "','"
                                + isSentValue + "','" + type + "')");
            }

        } finally {
            connection.close();
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.system.persistence.PersistenceService#deleteMessageExchange
     * (java.lang.String)
     */
    public void deleteMessageExchange(String exchangeId) throws SQLException {
        this.log.call("ExchangeId : " + exchangeId);
        final Connection connection = this.jdbcDataSource.getConnection();

        final Statement deleteStatement = connection.createStatement();
        try {
            deleteStatement.execute("DELETE FROM MESSAGE_EXCHANGE WHERE EXCHANGE_ID='" + exchangeId
                    + "'");
            deleteStatement.execute("DELETE FROM MESSAGE_EXCHANGE_PROPERTY WHERE EXCHANGE_ID='"
                    + exchangeId + "'");
            deleteStatement.execute("DELETE FROM NORMALIZED_MESSAGE WHERE EXCHANGE_ID='"
                    + exchangeId + "'");
            deleteStatement.execute("DELETE FROM EXCHANGE_TIMESTAMP WHERE EXCHANGE_ID='"
                    + exchangeId + "'");
        } finally {
            connection.close();
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.system.persistence.PersistenceService#getMessageExchange
     * (java.lang.String)
     */
    public MessageExchangeImpl getMessageExchange(String exchangeId) throws SQLException,
            IOException, ClassNotFoundException {
        this.log.call("ExchangeId : " + exchangeId);

        MessageExchangeImpl result = null;

        final Connection connection = this.jdbcDataSource.getConnection();

        final Statement getStatement = connection.createStatement();
        try {
            ResultSet resultSet = getStatement
                    .executeQuery("SELECT CONSUMER_ENDPOINT, ENDPOINT, INTERFACE_NAME, SERVICE_NAME, OPERATION_NAME, PATTERN, ERROR, ROLE, STATUS, TTERMINATED, TRANSACTED, MONITORED, PERSISTED FROM MESSAGE_EXCHANGE WHERE EXCHANGE_ID='"
                            + exchangeId + "'");

            if (resultSet.next()) {
                InputStream sqlInputStream;
                ObjectInputStream objectInputStream;
                final ServiceEndpoint consumerEndpoint;
                if (HSQLDB_DRIVER.equals(this.databaseDriver)) {
                    consumerEndpoint = (ServiceEndpoint) resultSet.getObject(1);
                } else {
                    sqlInputStream = resultSet.getBinaryStream(1);
                    objectInputStream = new ObjectInputStream(sqlInputStream);
                    consumerEndpoint = (ServiceEndpoint) objectInputStream.readObject();
                    objectInputStream.close();
                }

                final ServiceEndpoint endpoint;
                if (HSQLDB_DRIVER.equals(this.databaseDriver)) {
                    endpoint = (ServiceEndpoint) resultSet.getObject(2);
                } else {
                    sqlInputStream = resultSet.getBinaryStream(2);
                    if (sqlInputStream != null) {
                        objectInputStream = new ObjectInputStream(sqlInputStream);
                        endpoint = (ServiceEndpoint) objectInputStream.readObject();
                        objectInputStream.close();
                    } else {
                        endpoint = null;
                    }
                }

                final String interfaceNameAsString = resultSet.getString(3);
                QName interfaceName = null;
                if (interfaceNameAsString != null) {
                    interfaceName = QName.valueOf(interfaceNameAsString);
                }
                final String serviceAsString = resultSet.getString(4);
                QName service = null;
                if (serviceAsString != null) {
                    service = QName.valueOf(serviceAsString);
                }
                final String operationAsString = resultSet.getString(5);
                QName operation = null;
                if (operationAsString != null) {
                    operation = QName.valueOf(operationAsString);
                }
                final String pattern = resultSet.getString(6);

                final Exception error;
                if (HSQLDB_DRIVER.equals(this.databaseDriver)) {
                    error = (Exception) resultSet.getObject(7);
                } else {
                    sqlInputStream = resultSet.getBinaryStream(7);

                    if (sqlInputStream != null) {
                        objectInputStream = new ObjectInputStream(sqlInputStream);
                        error = (Exception) objectInputStream.readObject();
                        objectInputStream.close();
                    } else {
                        error = null;
                    }
                }

                short role = resultSet.getShort(8);
                short status = resultSet.getShort(9);
                final boolean terminated = resultSet.getBoolean(10);
                final boolean transacted = resultSet.getBoolean(11);
                final boolean monitored = resultSet.getBoolean(12);
                final boolean persisted = resultSet.getBoolean(13);

                final Map<String, Object> properties = new HashMap<String, Object>();
                resultSet = getStatement
                        .executeQuery("SELECT KKEY, VALUE FROM MESSAGE_EXCHANGE_PROPERTY WHERE EXCHANGE_ID='"
                                + exchangeId + "'");
                while (resultSet.next()) {
                    if (HSQLDB_DRIVER.equals(this.databaseDriver)) {
                        properties.put(resultSet.getString(1), resultSet.getObject(2));
                    } else {
                        sqlInputStream = resultSet.getBinaryStream(2);
                        if (sqlInputStream != null) {
                            objectInputStream = new ObjectInputStream(sqlInputStream);
                            properties.put(resultSet.getString(1), objectInputStream.readObject());
                            objectInputStream.close();
                        } else {
                            properties.put(resultSet.getString(1), null);
                        }
                    }
                }

                final Map<String, NormalizedMessage> messages = new HashMap<String, NormalizedMessage>();
                resultSet = getStatement
                        .executeQuery("SELECT KKEY, VALUE FROM NORMALIZED_MESSAGE WHERE EXCHANGE_ID='"
                                + exchangeId + "'");
                while (resultSet.next()) {
                    if (HSQLDB_DRIVER.equals(this.databaseDriver)) {
                        messages.put(resultSet.getString(1), (NormalizedMessage) resultSet
                                .getObject(2));
                    } else {
                        sqlInputStream = resultSet.getBinaryStream(2);
                        if (sqlInputStream != null) {
                            objectInputStream = new ObjectInputStream(sqlInputStream);
                            messages.put(resultSet.getString(1),
                                    (NormalizedMessage) objectInputStream.readObject());
                            objectInputStream.close();
                        } else {
                            messages.put(resultSet.getString(1), null);
                        }
                    }
                }

                try {
                    result = new MessageExchangeImpl(exchangeId, consumerEndpoint, endpoint,
                            interfaceName, service, operation, pattern, properties, messages,
                            error, role, status, terminated, transacted, monitored, persisted);
                } catch (URISyntaxException e) {
                    // Can not happend!
                }
            }
        } finally {
            connection.close();
        }

        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.system.persistence.PersistenceService#
     * getMessageExchangeHistory(java.lang.String)
     */
    public Map<String, Long> getMessageExchangeHistory(String exchangeId) throws SQLException {
        this.log.call("ExchangeId : " + exchangeId);

        Map<String, Long> result = null;

        final Connection connection = this.jdbcDataSource.getConnection();

        final Statement getStatement = connection.createStatement();
        try {
            ResultSet resultSet = getStatement
                    .executeQuery("SELECT EVENT_TIMESTAMP,IS_SENT,TYPE FROM EXCHANGE_TIMESTAMP WHERE EXCHANGE_ID='"
                            + exchangeId + "'");

            String suffix;
            while (resultSet.next()) {
                if (result == null) {
                    result = new HashMap<String, Long>();
                }
                if (resultSet.getBoolean(2)) {
                    suffix = "sent";
                } else {
                    suffix = "received";
                }
                result.put(resultSet.getString(3) + "-" + suffix, resultSet.getLong(1));
            }
        } finally {
            connection.close();
        }

        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.system.persistence.PersistenceService#getMessageExchangeIds
     * (javax.xml.namespace.QName, javax.xml.namespace.QName, java.lang.String,
     * javax.xml.namespace.QName, long, long)
     */
    public List<String> getMessageExchangeIds(QName interfaceName, QName serviceName,
            String endpointName, QName operationName, long startDate, long endDate)
            throws SQLException {
        this.log.call();

        List<String> result = new ArrayList<String>();
        final Connection connection = this.jdbcDataSource.getConnection();

        final Statement getStatement = connection.createStatement();

        StringBuilder conditionStatementBuilder = new StringBuilder();
        if (interfaceName != null) {
            conditionStatementBuilder.append("INTERFACE_NAME='" + interfaceName.toString()
                    + "' AND ");
        }
        if (serviceName != null) {
            conditionStatementBuilder.append("SERVICE_NAME='" + serviceName.toString() + "' AND ");
        }
        if (endpointName != null) {
            conditionStatementBuilder.append("ENDPOINT_NAME='" + endpointName + "' AND ");
        }
        if (operationName != null) {
            conditionStatementBuilder.append("OPERATION_NAME='" + operationName.toString()
                    + "' AND ");
        }
        try {
            final ResultSet resultSet = getStatement
                    .executeQuery("SELECT DISTINCT MESSAGE_EXCHANGE.EXCHANGE_ID FROM MESSAGE_EXCHANGE,EXCHANGE_TIMESTAMP WHERE "
                            + conditionStatementBuilder.toString()
                            + "EXCHANGE_TIMESTAMP.EXCHANGE_ID=MESSAGE_EXCHANGE.EXCHANGE_ID AND EXCHANGE_TIMESTAMP.EVENT_TIMESTAMP BETWEEN '"
                            + startDate + "' AND '" + endDate + "'");
            while (resultSet.next()) {
                result.add(resultSet.getString(1));
            }
        } finally {
            connection.close();
        }
        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.system.persistence.PersistenceService#getMessageExchanges
     * (javax.xml.namespace.QName, javax.xml.namespace.QName, java.lang.String,
     * javax.xml.namespace.QName, long, long)
     */
    public int getMessageExchanges(QName interfaceName, QName serviceName, String endpointName,
            QName operationName, long startDate, long endDate) throws SQLException {
        this.log.call();

        int result = 0;
        final Connection connection = this.jdbcDataSource.getConnection();

        final Statement getStatement = connection.createStatement();

        StringBuilder conditionStatementBuilder = new StringBuilder();
        if (interfaceName != null) {
            conditionStatementBuilder.append("MESSAGE_EXCHANGE.INTERFACE_NAME='"
                    + interfaceName.toString() + "' AND ");
        }
        if (serviceName != null) {
            conditionStatementBuilder.append("MESSAGE_EXCHANGE.SERVICE_NAME='"
                    + serviceName.toString() + "' AND ");
        }
        if (endpointName != null) {
            conditionStatementBuilder.append("MESSAGE_EXCHANGE.ENDPOINT_NAME='" + endpointName
                    + "' AND ");
        }
        if (operationName != null) {
            conditionStatementBuilder.append("MESSAGE_EXCHANGE.OPERATION_NAME='"
                    + operationName.toString() + "' AND ");
        }
        try {
            final ResultSet resultSet = getStatement
                    .executeQuery("SELECT COUNT(DISTINCT MESSAGE_EXCHANGE.EXCHANGE_ID) FROM MESSAGE_EXCHANGE,EXCHANGE_TIMESTAMP WHERE "
                            + conditionStatementBuilder.toString()
                            + " MESSAGE_EXCHANGE.EXCHANGE_ID=EXCHANGE_TIMESTAMP.EXCHANGE_ID AND EXCHANGE_TIMESTAMP.EVENT_TIMESTAMP BETWEEN '"
                            + startDate + "' AND '" + endDate + "'");
            if (resultSet.next()) {
                result = resultSet.getInt(1);
            }
        } finally {
            connection.close();
        }

        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.system.persistence.PersistenceService#
     * setExchangeStorageDuration(long)
     */
    public void setExchangeStorageDuration(long duration) {
        this.storageDuration = duration;
        this.cleanerTimer.cancel();
        this.cleanerTimer = new Timer();
        this.cleanerTimer.schedule(new PersistenceCleaner(), 0, this.storageDuration);

    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.system.persistence.PersistenceService#updateMessageExchange
     * (org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl, boolean,
     * java.lang.Boolean)
     */
    public void updateMessageExchange(MessageExchangeImpl exchange, boolean storeExchange,
            Boolean isSent) throws SQLException, IOException {
        this.log.call("ExchangeId : " + exchange.getExchangeId());

        final Connection connection = this.jdbcDataSource.getConnection();

        final Statement addStatement = connection.createStatement();

        try {
            if (!storeExchange) {
                if (!ExchangeStatus.ACTIVE.equals(exchange.getStatus())) {
                    short status;
                    if (ExchangeStatus.DONE.equals(exchange.getStatus())) {
                        status = MessageExchangeImpl.SERIALIZE_STATUS_DONE;
                    } else {
                        status = MessageExchangeImpl.SERIALIZE_STATUS_ERROR;
                    }
                    connection.createStatement().execute(
                            "UPDATE MESSAGE_EXCHANGE SET STATUS ='" + status
                                    + "' WHERE EXCHANGE_ID='" + exchange.getExchangeId() + "'");
                }
            } else {
                PreparedStatement storeStatement = connection
                        .prepareStatement("UPDATE MESSAGE_EXCHANGE SET CONSUMER_ENDPOINT=?, ENDPOINT=?, INTERFACE_NAME=?, SERVICE_NAME=?, ENDPOINT_NAME=?, ERROR=?, ROLE=?, STATUS=?, TTERMINATED=?, TRANSACTED=?, MONITORED=?, PERSISTED=? WHERE EXCHANGE_ID=?");

                storeStatement.setObject(1, exchange.getConsumerEndpoint());
                storeStatement.setObject(2, exchange.getServiceEndpoint());
                storeStatement.setString(3, exchange.getInterfaceName().toString());
                storeStatement.setString(4, exchange.getService().toString());
                storeStatement.setString(5, exchange.getServiceEndpoint().getEndpointName());
                storeStatement.setObject(6, exchange.getError());
                if (Role.CONSUMER.equals(exchange.getRole())) {
                    storeStatement.setShort(7, MessageExchangeImpl.SERIALIZE_ROLE_CONSUMER);
                } else {
                    storeStatement.setShort(7, MessageExchangeImpl.SERIALIZE_ROLE_PROVIDER);
                }
                if (ExchangeStatus.ACTIVE.equals(exchange.getStatus())) {
                    storeStatement.setShort(8, MessageExchangeImpl.SERIALIZE_STATUS_ACTIVE);
                } else if (ExchangeStatus.DONE.equals(exchange.getStatus())) {
                    storeStatement.setShort(8, MessageExchangeImpl.SERIALIZE_STATUS_DONE);
                } else {
                    storeStatement.setShort(8, MessageExchangeImpl.SERIALIZE_STATUS_ERROR);
                }
                storeStatement.setBoolean(9, exchange.isTerminated());
                storeStatement.setBoolean(10, exchange.isTransacted());
                storeStatement.setBoolean(11, exchange.isMonitored());
                storeStatement.setBoolean(12, exchange.isPersisted());
                storeStatement.setString(13, exchange.getExchangeId());

                storeStatement.execute();

                storeStatement = connection
                        .prepareStatement("INSERT INTO MESSAGE_EXCHANGE_PROPERTY VALUES(?,?,?,?)");

                for (Object key : exchange.getPropertyNames()) {
                    storeStatement.setString(1, exchange.getExchangeId() + (String) key);
                    storeStatement.setString(2, exchange.getExchangeId());
                    storeStatement.setString(3, (String) key);
                    if (HSQLDB_DRIVER.equals(this.databaseDriver)) {
                        storeStatement.setObject(4, exchange.getProperty((String) key));
                    } else {
                        // With MYSQL, the setObject() method serializes not
                        // properly primitive alter-ego Objects, use Streams
                        // instead
                        final ByteArrayOutputStream bytesOutputStream = new ByteArrayOutputStream();
                        new ObjectOutputStream(bytesOutputStream).writeUnshared(exchange
                                .getProperty((String) key));
                        ByteArrayInputStream bytesInputStream = new ByteArrayInputStream(
                                bytesOutputStream.toByteArray());
                        storeStatement.setBinaryStream(4, bytesInputStream, bytesInputStream
                                .available());
                    }

                    try {
                        storeStatement.execute();
                    } catch (SQLException e) {
                        // the entry may already exist, do nothing
                    }
                }

                storeStatement = connection
                        .prepareStatement("INSERT INTO NORMALIZED_MESSAGE VALUES(?,?,?,?)");
                for (String key : exchange.getMessages().keySet()) {
                    storeStatement.setString(1, exchange.getExchangeId() + key);
                    storeStatement.setString(2, exchange.getExchangeId());
                    storeStatement.setString(3, key);
                    storeStatement.setObject(4, exchange.getMessage(key));

                    try {
                        storeStatement.execute();
                    } catch (SQLException e) {
                        // the entry may already exist, do nothing
                    }
                }
            }

            if (isSent != null) {
                short isSentbinary = 0;
                if (isSent) {
                    isSentbinary = 1;
                }
                final String type;
                if (exchange.getMessage(MessageExchangeImpl.IN_MSG) != null) {
                    type = MessageExchangeImpl.IN_MSG;
                } else if (exchange.getMessage(MessageExchangeImpl.OUT_MSG) != null) {
                    type = MessageExchangeImpl.OUT_MSG;
                } else if (exchange.getMessage(MessageExchangeImpl.FAULT_MSG) != null) {
                    type = MessageExchangeImpl.FAULT_MSG;
                } else if (exchange.getError() != null) {
                    type = "error-ack";
                } else {
                    type = "done-ack";
                }
                
                String isSentValue = null;
                /*
                 * If Hsqldb use 'true'/'false' values
                 * for is_sent column, else use
                 * '1'/'0' values
                 */
                if(HSQLDB_DRIVER.equals(this.databaseDriver)){
                    if(isSentbinary == 0){
                        isSentValue = "false";
                    } else {
                        isSentValue = "true";
                    }
                } else {
                    isSentValue = String.valueOf(isSentbinary);
                }              
                
                addStatement
                        .execute("INSERT INTO EXCHANGE_TIMESTAMP (EVENT_TIMESTAMP, EXCHANGE_ID, IS_SENT, TYPE) VALUES('"
                                + System.currentTimeMillis()
                                + "','"
                                + exchange.getExchangeId()
                                + "','" + isSentValue + "','" + type + "')");
            }
        } finally {
            connection.close();
        }

    }

    @LifeCycle(on = LifeCycleType.START)
    protected void start() throws SQLException {
        this.log = new LoggingUtil(this.logger);
        this.log.start();

        final ContainerConfiguration containerConfiguration = this.configurationService
                .getContainerConfiguration();
        this.storageDuration = containerConfiguration.getPersistenceDuration();
        this.databaseDriver = containerConfiguration.getRegistryDatabaseDriver();

        this.jdbcDataSource = new BasicDataSource();
        this.jdbcDataSource.setDriverClassName(containerConfiguration.getRegistryDatabaseDriver());
        this.jdbcDataSource.setUrl(containerConfiguration.getRegistryDatabaseUrl());
        this.jdbcDataSource.setUsername(containerConfiguration.getRegistryDatabaseUser());
        this.jdbcDataSource.setPassword(containerConfiguration.getRegistryDatabasePassword());

        try {
            this.checkPersistenceTables(this.jdbcDataSource.getConnection());
        } catch (final SQLException e) {
            this.log.error("Failed to start persistence service", e);
            throw e;
        }

        this.cleanerTimer = new Timer();
        this.cleanerTimer.schedule(new PersistenceCleaner(), storageDuration, storageDuration);

        this.log.end();
    }

    @LifeCycle(on = LifeCycleType.STOP)
    protected void stop() {
        this.log.call();

        this.cleanerTimer.cancel();

        try {
            this.jdbcDataSource.close();
        } catch (final SQLException e) {
            this.log.debug(e);
        }
    }

    /**
     * Check id the tables exists. <br/>
     * If not the method create the persistence tables.
     * 
     * @param connection
     * @throws SQLException
     */
    private final void checkPersistenceTables(final Connection connection) throws SQLException {

        final Statement statementCheck = connection.createStatement();
        boolean isTablesCreated;

        try {
            statementCheck.executeQuery("SELECT 1 FROM MESSAGE_EXCHANGE");
            isTablesCreated = true;
        } catch (final SQLException e) {
            isTablesCreated = false;
        } finally {
            statementCheck.close();
        }

        if (!isTablesCreated) {
            this.log.info("Creating database tables");

            final Statement statementCreate = connection.createStatement();
            String stringStatement;

            // create the Message Exchange table
            if (HSQLDB_DRIVER.equals(databaseDriver)) {
                statementCreate.executeUpdate("SET WRITE_DELAY FALSE");
                stringStatement = "CREATE TABLE MESSAGE_EXCHANGE (EXCHANGE_ID VARCHAR("
                        + IDGenerator.UUID_SIZE
                        + ") PRIMARY KEY, CONSUMER_ENDPOINT OBJECT, ENDPOINT OBJECT, INTERFACE_NAME VARCHAR(255), SERVICE_NAME VARCHAR(255), ENDPOINT_NAME VARCHAR(255), OPERATION_NAME VARCHAR(255), PATTERN VARCHAR(63), ERROR OBJECT, ROLE TINYINT, STATUS TINYINT, TTERMINATED BOOLEAN, TRANSACTED BOOLEAN, MONITORED BOOLEAN, PERSISTED BOOLEAN, CREATION_TIMESTAMP BIGINT)";
            } else if (ORACLE_DRIVER.equals(databaseDriver)) {           	
                stringStatement = "CREATE TABLE MESSAGE_EXCHANGE (EXCHANGE_ID VARCHAR("
                        + IDGenerator.UUID_SIZE
                        + ") PRIMARY KEY, CONSUMER_ENDPOINT BLOB, ENDPOINT BLOB, INTERFACE_NAME VARCHAR(255), SERVICE_NAME VARCHAR(255), ENDPOINT_NAME VARCHAR(255), OPERATION_NAME VARCHAR(255), PATTERN VARCHAR(63), ERROR BLOB, ROLE NUMBER(3), STATUS NUMBER(3), TTERMINATED NUMBER(1), TRANSACTED NUMBER(1), MONITORED NUMBER(1), PERSISTED NUMBER(1), CREATION_TIMESTAMP NUMBER(38))";
            }else {
                stringStatement = "CREATE TABLE MESSAGE_EXCHANGE (EXCHANGE_ID VARCHAR("
                        + IDGenerator.UUID_SIZE
                        + ") PRIMARY KEY, CONSUMER_ENDPOINT LONGBLOB, ENDPOINT LONGBLOB, INTERFACE_NAME VARCHAR(255), SERVICE_NAME VARCHAR(255), ENDPOINT_NAME VARCHAR(255), OPERATION_NAME VARCHAR(255), PATTERN VARCHAR(63), ERROR LONGBLOB, ROLE TINYINT, STATUS TINYINT, TTERMINATED BOOLEAN, TRANSACTED BOOLEAN, MONITORED BOOLEAN, PERSISTED BOOLEAN, CREATION_TIMESTAMP BIGINT)";
            }
            statementCreate.executeUpdate(stringStatement);
            stringStatement = "CREATE INDEX INTERFACE_INDEX ON MESSAGE_EXCHANGE (INTERFACE_NAME)";
            statementCreate.executeUpdate(stringStatement);
            stringStatement = "CREATE INDEX SERVICE_INDEX ON MESSAGE_EXCHANGE (SERVICE_NAME)";
            statementCreate.executeUpdate(stringStatement);
            stringStatement = "CREATE INDEX ENDPOINT_INDEX ON MESSAGE_EXCHANGE (ENDPOINT_NAME)";
            statementCreate.executeUpdate(stringStatement);
            stringStatement = "CREATE INDEX OPERATION_INDEX ON MESSAGE_EXCHANGE (OPERATION_NAME)";

            // create the Message Exchange Property table
            if (HSQLDB_DRIVER.equals(databaseDriver)) {
                stringStatement = "CREATE TABLE MESSAGE_EXCHANGE_PROPERTY (ID VARCHAR(255) PRIMARY KEY, EXCHANGE_ID VARCHAR("
                        + IDGenerator.UUID_SIZE + ") NOT NULL, KKEY VARCHAR(255), VALUE OBJECT)";
            } else if (ORACLE_DRIVER.equals(databaseDriver)) {            	
            	stringStatement = "CREATE TABLE MESSAGE_EXCHANGE_PROPERTY (ID VARCHAR(255) PRIMARY KEY, EXCHANGE_ID VARCHAR("
                    + IDGenerator.UUID_SIZE + ") NOT NULL, KKEY VARCHAR(255), VALUE BLOB)";
            }else {
                stringStatement = "CREATE TABLE MESSAGE_EXCHANGE_PROPERTY (ID VARCHAR(255) PRIMARY KEY, EXCHANGE_ID VARCHAR("
                        + IDGenerator.UUID_SIZE + ") NOT NULL, KKEY VARCHAR(255), VALUE LONGBLOB)";
            }
            statementCreate.executeUpdate(stringStatement);
            statementCreate
                    .executeUpdate("ALTER TABLE MESSAGE_EXCHANGE_PROPERTY ADD CONSTRAINT M_E_PROPERTY_FOREIGN_KEY FOREIGN KEY (EXCHANGE_ID) REFERENCES MESSAGE_EXCHANGE(EXCHANGE_ID)");

            // create the Normalized Message table
            if (HSQLDB_DRIVER.equals(databaseDriver)) {
                stringStatement = "CREATE TABLE NORMALIZED_MESSAGE (ID VARCHAR(255) PRIMARY KEY, EXCHANGE_ID VARCHAR("
                        + IDGenerator.UUID_SIZE + ") NOT NULL, KKEY VARCHAR(5), VALUE OBJECT)";
            } else if (ORACLE_DRIVER.equals(databaseDriver)) {
            	stringStatement = "CREATE TABLE NORMALIZED_MESSAGE (ID VARCHAR(255) PRIMARY KEY, EXCHANGE_ID VARCHAR("
                    + IDGenerator.UUID_SIZE + ") NOT NULL, KKEY VARCHAR(5), VALUE BLOB)";
            }else {
                stringStatement = "CREATE TABLE NORMALIZED_MESSAGE (ID VARCHAR(255) PRIMARY KEY, EXCHANGE_ID VARCHAR("
                        + IDGenerator.UUID_SIZE + ") NOT NULL, KKEY VARCHAR(5), VALUE LONGBLOB)";
            }
            statementCreate.executeUpdate(stringStatement);
            statementCreate
                    .executeUpdate("ALTER TABLE NORMALIZED_MESSAGE ADD CONSTRAINT N_M_FOREIGN_KEY FOREIGN KEY (EXCHANGE_ID) REFERENCES MESSAGE_EXCHANGE(EXCHANGE_ID)");

            // create exchange send/receive date table
            if (HSQLDB_DRIVER.equals(databaseDriver)) {
                stringStatement = "CREATE TABLE EXCHANGE_TIMESTAMP (ID IDENTITY, EVENT_TIMESTAMP BIGINT, EXCHANGE_ID VARCHAR("
                        + IDGenerator.UUID_SIZE + ") NOT NULL, IS_SENT BOOLEAN, TYPE VARCHAR(5))";
            } else if (ORACLE_DRIVER.equals(databaseDriver)) {
            	stringStatement = "CREATE TABLE EXCHANGE_TIMESTAMP (ID NUMBER(38), EVENT_TIMESTAMP NUMBER(38), EXCHANGE_ID VARCHAR("
                    + IDGenerator.UUID_SIZE + ") NOT NULL, IS_SENT NUMBER(1), TYPE VARCHAR(5))";
            }else {
                stringStatement = "CREATE TABLE EXCHANGE_TIMESTAMP (ID MEDIUMINT NOT NULL AUTO_INCREMENT PRIMARY KEY, EVENT_TIMESTAMP BIGINT, EXCHANGE_ID VARCHAR("
                        + IDGenerator.UUID_SIZE + ") NOT NULL, IS_SENT BOOLEAN, TYPE VARCHAR(5))";
            }
            statementCreate.executeUpdate(stringStatement);
            statementCreate
                    .executeUpdate("ALTER TABLE EXCHANGE_TIMESTAMP ADD CONSTRAINT E_T_FOREIGN_KEY FOREIGN KEY (EXCHANGE_ID) REFERENCES MESSAGE_EXCHANGE(EXCHANGE_ID)");
        }
    }
}
